#!/usr/bin/env python

# Author: Trevor Sherrard
# Project: Styx LED Controller 
# Since: March 1st, 2021
# Description: This python script interacts with a LED load using the Styx LED controller
#              controlled via an i2c interface. This generates color values with different
#              intensity values based on the results of an fft from data retrieved from an
#              audio input device.

import numpy as np
import pyaudio
import requests
import smbus
import time

class SpectrumAnalyzer:
    # create constants that will be used throughout the developed software.
    
    # constants for audio input and FFT
    FORMAT = pyaudio.paFloat32
    CHANNELS = 1
    CHUNK = 512
    START = 0
    N = 512
    RATE = 44100

    # CHANGE THIS TO YOUR DEVICE INDEX
    device_index = 2
    
    # i2c constants
    holdCmd = 'h'
    i2c_channel = 1
    i2c_address = 0x04

    # place holders for raw data and fft results
    wave_x = 0
    wave_y = 0
    spec_x = 0
    spec_y = 0
    data = []

    # constants for scaling and data conditioning
    lowDivisor = 3
    medDivisor = 5
    highDivisor = 5
    multiplier = 255
    lowToMedThresh = 1300
    medToHighThresh = 3000
    upperHigh = 7000
    
    def __init__(self):
        # create audio stream object instance
        self.pa = pyaudio.PyAudio()
        self.stream = self.pa.open(format = self.FORMAT,
            channels = self.CHANNELS, 
            rate = self.RATE, 
            input = True,
            output = False,
            frames_per_buffer = self.CHUNK,
            input_device_index = self.device_index)

        # create i2c bus connection
        self.bus = smbus.SMBus(self.i2c_channel)
    
        # Main loop
        self.loop()

    def valueToHexString(self, val):
        """
        this function returns a two hex digit string representation
        of a decimal number between 0 and 255. 

        params:
            self: current object instance
            val (int): decimal number between 0 and 255
        returns:
            hexString (str): The string representation of the hex value
        """
        hexVal = hex(int(val))[2:]
        hexString = str(hexVal).upper()
        
        # pad with zeros if nessesary
        if(len(hexString) < 2):
            hexString = "0" + hexString
        return hexString

    def loop(self):
        try:
            while True :
                # update data buffer with new audio
                self.data = self.audioinput()

                # perform fft
                self.fft()
                
                # map values
                mappedLow = ((self.lowAvg)/self.lowDivisor)*self.multiplier
                mappedMid = ((self.medAvg)/self.medDivisor)*self.multiplier
                mappedHigh = ((self.highAvg)/self.highDivisor)*self.multiplier
                
                # create dictionary 
                mappedDict = {"low":mappedLow, "med":mappedMid, "high":mappedHigh}

                # find frequency bin with max average value
                maxKey = max(mappedDict, key=mappedDict.get)
                
                # set color based on key
                maxVal = mappedDict[maxKey]

                # cap out max Val at 255
                if(maxVal > 255):
                    maxVal = 255

                # construct hex string from scaled average value
                # from the frequency bin with the maximum value
                hexString = self.valueToHexString(maxVal)
                colorVal = ""

                # red for low frequency values
                if(maxKey == "low"):
                    colorVal = hexString + "000000"

                # green for medium frequency values
                elif(maxKey == "med"):
                    colorVal = "00" + hexString + "0000"

                # blue for high frequency values
                elif(maxKey == "high"):
                    colorVal = "0000" + hexString + "00"

                # write color to lights
                self.sendString(self.holdCmd, colorVal, 0.05)

        except KeyboardInterrupt:
            self.pa.close()

    def audioinput(self):
        """
        This function reads an audio sample of size CHUNK from the audio input device
        and returns said data as a numpy array.

        params:
            self: Current object instance
        returns:
            ret (numpy array): The audio data from the input device
        """
        ret = self.stream.read(self.CHUNK, exception_on_overflow = False)
        ret = np.fromstring(ret, np.float32)
        return ret

    def fft(self):
        """
        This function performs the FFT on an acquired chunk of audio data, and exctracts the average 
        spectral instensity over three static frequency bins. These values are updated within the current
        object instance.

        params:
            self: current object instance.
        returns:
            None
        """

        # get get discretized x-axis scaling
        self.wave_x = range(self.START, self.START + self.N)
        
        # get amplitude data for each sample instance
        self.wave_y = self.data[self.START:self.START + self.N]
        
        # get individual frequency domain x-axis values
        self.spec_x = np.fft.fftfreq(self.N, d = 1.0 / self.RATE)  
        
        # perform actual fft
        y = np.fft.fft(self.data[self.START:self.START + self.N])    
        
        # get spectral magnitude from complex fft result values.
        self.spec_y = [np.sqrt(c.real ** 2 + c.imag ** 2) for c in y]

        # perform frequency binning (low, medium, high)
        low_freq = [self.spec_y[i] for i in range(len(self.spec_x)) if self.spec_x[i] < self.lowToMedThresh]
        med_freq = [self.spec_y[i] for i in range(len(self.spec_x)) if (self.spec_x[i] > self.lowToMedThresh and self.spec_x[i] < self.medToHighThresh)]
        high_freq = [self.spec_y[i] for i in range(len(self.spec_x)) if (self.spec_x[i] > self.medToHighThresh and self.spec_x[i] < self.upperHigh)]
        
        # get average of each bin
        self.lowAvg = np.mean(low_freq)
        self.medAvg = np.mean(med_freq)
        self.highAvg = np.mean(high_freq)

    def sendString(self, commandString, colorString, delayVal):
        """
        This function converts the stringToSend to bytes and then sends
        these bytes via the i2c interface to the LED controller.

        params:
            commandString (str): the given command for the given color.
            colorString (str): the desired color to change the LEDs to.
            bus (smbus object instance): The smbus object instance used for i2c communication
        returns:
            None
        """
        # convert string to bytes
        byteList = []
        commandByte = ord(commandString)
        for char in colorString:
            byteVal = ord(char)
            byteList.append(byteVal)

        # send data block
        self.bus.write_i2c_block_data(self.i2c_address, commandByte, byteList)

        # add settling time for i2c
        time.sleep(delayVal)

if __name__ == "__main__":
    spec = SpectrumAnalyzer()
    
