#!/usr/bin/env python

# Author: Trevor Sherrard
# Project: Styx Networked LED controller
# Since: 02/24/2021
# Description: This python script provides functionallity to communicate
#              with the Styx LED controller via an i2c interface and provides examples of how to 
#              change the state of the LED controller via said interface

import smbus
import time

# define constants
i2c_channel = 1
i2c_address = 0x04
delay = 5
flash_delay = 20
ramp_delay = 30

# define commands
holdCmd = 'h'
flashCmd = 'f'
rampCmd = 'r'
strobeCmd = 's'

# define colors
redStr = "FF000000"
blueStr = "0000FF00"
whiteStr = "FFFF4000"

# define flash amount and fade delays
NumFlashString = "05" # note the left padding for values less than 10
NumFadeInc = "50" # this represents 50, 100ms intervals (5 seconds)

def sendString(commandString, stringToSend, bus):
    """
    This function converts the stringToSend to bytes and then sends 
    these bytes via the i2c interface to the LED controller. 

    params:
        commandString (str): The command character for each board function (r -> ramp, h -> hold,
            f -> flash, s -> strobe, c -> clear)
        stringToSend (str): The string to be sent via the i2c interface (color and duration information (if applicable))
        bus (smbus object instance): The smbus object instance used for i2c communication
    returns:
        None
    """

    # convert string to bytes
    byteList = []
    commandByte = ord(commandString)
    for char in stringToSend:
        byteVal = ord(char)
        byteList.append(byteVal)

    # send data block
    bus.write_i2c_block_data(i2c_address, commandByte, byteList)

def main():
    bus = smbus.SMBus(i2c_channel)

    # first demonstrate the hold command
    sendString(holdCmd, redStr, bus)
    time.sleep(delay)

    sendString(holdCmd, whiteStr, bus)
    time.sleep(delay)

    sendString(holdCmd, blueStr, bus)
    time.sleep(delay)

    # then, demonstrate the flash command
    redFlashString = redStr + ":" + NumFlashString
    sendString(flashCmd, redFlashString, bus)
    time.sleep(flash_delay)

    whiteFlashString = whiteStr + ":" + NumFlashString
    sendString(flashCmd, whiteFlashString, bus)
    time.sleep(flash_delay)

    blueFlashString = blueStr + ":" + NumFlashString
    sendString(flashCmd, blueFlashString, bus)
    time.sleep(flash_delay)

    # then, demonstrate the ramp command
    redRampString = redStr + ":" + NumFadeInc
    sendString(rampCmd, redRampString, bus)
    time.sleep(ramp_delay)

    whiteRampString = whiteStr + ":" + NumFadeInc
    sendString(rampCmd, whiteRampString, bus)
    time.sleep(ramp_delay)

    blueRampString = blueStr + ":" + NumFadeInc
    sendString(rampCmd, blueRampString, bus)
    time.sleep(ramp_delay)

    # then, demonstrate the strobe command
    sendString(strobeCmd, redStr, bus)
    time.sleep(delay)

    sendString(strobeCmd, whiteStr, bus)
    time.sleep(delay)

    sendString(strobeCmd, blueStr, bus)
    time.sleep(delay)

if(__name__ == "__main__"):
    main()
